@@ -4,8 +4,12 @@
 
     This program can be distributed under the terms of the GNU GPL.
     See the file COPYING.
+
+    gcc -Wall `pkg-config fuse --cflags --libs` null.c -o null
 */
 
+#define FUSE_USE_VERSION 26
+
 #include <fuse.h>
 #include <string.h>
 #include <unistd.h>
@@ -84,5 +88,5 @@ static struct fuse_operations null_oper = {
 
 int main(int argc, char *argv[])
 {
-    return fuse_main(argc, argv, &null_oper);
+    return fuse_main(argc, argv, &null_oper, NULL);
 }
