@@ -7,11 +7,14 @@
 */
 
 #include "fuse_lowlevel.h"
+#include "fuse_common_compat.h"
+#include "fuse_lowlevel_compat.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
+#include <errno.h>
 
 struct fuse_session {
     struct fuse_session_ops op;
@@ -33,6 +36,8 @@ struct fuse_chan {
     size_t bufsize;
 
     void *data;
+
+    int compat;
 };
 
 struct fuse_session *fuse_session_new(struct fuse_session_ops *op, void *data)
@@ -58,6 +63,16 @@ void fuse_session_add_chan(struct fuse_session *se, struct fuse_chan *ch)
     ch->se = se;
 }
 
+void fuse_session_remove_chan(struct fuse_chan *ch)
+{
+    struct fuse_session *se = ch->se;
+    if (se) {
+        assert(se->ch == ch);
+        se->ch = NULL;
+        ch->se = NULL;
+    }
+}
+
 struct fuse_chan *fuse_session_next_chan(struct fuse_session *se,
                                          struct fuse_chan *ch)
 {
@@ -105,8 +120,8 @@ int fuse_session_exited(struct fuse_session *se)
         return se->exited;
 }
 
-struct fuse_chan *fuse_chan_new(struct fuse_chan_ops *op, int fd,
-                                size_t bufsize, void *data)
+static struct fuse_chan *fuse_chan_new_common(struct fuse_chan_ops *op, int fd,
+                                size_t bufsize, void *data, int compat)
 {
     struct fuse_chan *ch = (struct fuse_chan *) malloc(sizeof(*ch));
     if (ch == NULL) {
@@ -119,10 +134,24 @@ struct fuse_chan *fuse_chan_new(struct fuse_chan_ops *op, int fd,
     ch->fd = fd;
     ch->bufsize = bufsize;
     ch->data = data;
+    ch->compat = compat;
 
     return ch;
 }
 
+struct fuse_chan *fuse_chan_new(struct fuse_chan_ops *op, int fd,
+                                size_t bufsize, void *data)
+{
+    return fuse_chan_new_common(op, fd, bufsize, data, 0);
+}
+
+struct fuse_chan *fuse_chan_new_compat24(struct fuse_chan_ops_compat24 *op,
+                                         int fd, size_t bufsize, void *data)
+{
+    return fuse_chan_new_common((struct fuse_chan_ops *) op, fd, bufsize,
+                                data, 24);
+}
+
 int fuse_chan_fd(struct fuse_chan *ch)
 {
     return ch->fd;
@@ -143,9 +172,22 @@ struct fuse_session *fuse_chan_session(struct fuse_chan *ch)
     return ch->se;
 }
 
+int fuse_chan_recv(struct fuse_chan **chp, char *buf, size_t size)
+{
+    struct fuse_chan *ch = *chp;
+    if (ch->compat)
+        return ((struct fuse_chan_ops_compat24 *) &ch->op)
+            ->receive(ch, buf, size);
+    else
+        return ch->op.receive(chp, buf, size);
+}
+
 int fuse_chan_receive(struct fuse_chan *ch, char *buf, size_t size)
 {
-    return ch->op.receive(ch, buf, size);
+    int res;
+
+    res = fuse_chan_recv(&ch, buf, size);
+    return res >= 0 ? res : (res != -EINTR && res != -EAGAIN) ? -1 : 0;
 }
 
 int fuse_chan_send(struct fuse_chan *ch, const struct iovec iov[], size_t count)
@@ -155,7 +197,12 @@ int fuse_chan_send(struct fuse_chan *ch, const struct iovec iov[], size_t count)
 
 void fuse_chan_destroy(struct fuse_chan *ch)
 {
+    fuse_session_remove_chan(ch);
     if (ch->op.destroy)
         ch->op.destroy(ch);
     free(ch);
 }
+
+#ifndef __FreeBSD__
+__asm__(".symver fuse_chan_new_compat24,fuse_chan_new@FUSE_2.4");
+#endif
