@@ -39,6 +39,10 @@
 #include <sys/utsname.h>
 #include <sys/sysmacros.h>
 
+#ifdef HAVE_SELINUX_SELINUX_H
+#include <selinux/selinux.h>
+#endif
+
 #define FUSE_COMMFD_ENV         "_FUSE_COMMFD"
 
 #define FUSE_DEV_OLD "/proc/fs/fuse/dev"
@@ -104,6 +108,13 @@ static int lock_mtab(void)
     const char *mtab_lock = _PATH_MOUNTED ".fuselock";
     int mtablock;
     int res;
+    struct stat mtab_stat;
+
+    /* /etc/mtab could be a symlink to /proc/mounts */
+    if (!lstat(_PATH_MOUNTED, &mtab_stat)) {
+        if (S_ISLNK(mtab_stat.st_mode))
+            return -1;
+    }
 
     mtablock = open(mtab_lock, O_RDWR | O_CREAT, 0600);
     if (mtablock >= 0) {
@@ -130,7 +141,7 @@ static void unlock_mtab(int mtablock)
    filenames. */
 static int check_name(const char *name)
 {
-    char *s;
+    const char *s;
     for (s = "\n\t\\"; *s; s++) {
         if (strchr(name, *s)) {
             fprintf(stderr, "%s: illegal character 0x%02x in mount entry\n",
@@ -185,6 +196,18 @@ static int unmount_rename(const char *mtab, const char *mtab_new)
     if (stat(mtab, &sbuf) == 0)
         chown(mtab_new, sbuf.st_uid, sbuf.st_gid);
 
+#ifdef HAVE_LIBSELINUX
+    {
+        security_context_t filecon;
+
+        if (getfilecon(mtab, &filecon) > 0) {
+            setfilecon(mtab_new, filecon);
+            if (filecon != NULL)
+                freecon(filecon);
+        }
+    }
+#endif
+
     res = rename(mtab_new, mtab);
     if (res == -1) {
         fprintf(stderr, "%s: failed to rename %s to %s: %s\n", progname,
@@ -241,7 +264,8 @@ static int unmount_fuse(const char *mnt, int quiet, int lazy)
     while ((entp = getmntent(fp)) != NULL) {
         int removed = 0;
         if (!found && strcmp(entp->mnt_dir, mnt) == 0 &&
-           strcmp(entp->mnt_type, "fuse") == 0) {
+            (strcmp(entp->mnt_type, "fuse") == 0 ||
+             strcmp(entp->mnt_type, "fuseblk") == 0)) {
             if (user == NULL)
                 removed = 1;
             else {
@@ -528,7 +552,7 @@ static int check_mountpoint_empty(const char *mnt, mode_t rootmode,
         struct dirent *ent;
         DIR *dp = opendir(mnt);
         if (dp == NULL) {
-            fprintf(stderr, "%s: failed to mountpoint for reading: %s\n",
+            fprintf(stderr, "%s: failed to open mountpoint for reading: %s\n",
                     progname, strerror(errno));
             return -1;
         }
@@ -551,7 +575,24 @@ static int check_mountpoint_empty(const char *mnt, mode_t rootmode,
     return 0;
 }
 
-static int do_mount(const char *mnt, const char *type, mode_t rootmode,
+static int has_fuseblk(void)
+{
+    char buf[256];
+    FILE *f = fopen("/proc/filesystems", "r");
+    if (!f)
+        return 1;
+
+    while (fgets(buf, sizeof(buf), f))
+        if (strstr(buf, "fuseblk\n")) {
+            fclose(f);
+            return 1;
+        }
+
+    fclose(f);
+    return 0;
+}
+
+static int do_mount(const char *mnt, const char **type, mode_t rootmode,
                     int fd, const char *opts, const char *dev, char **fsnamep,
                     char **mnt_optsp, off_t rootsize)
 {
@@ -563,6 +604,7 @@ static int do_mount(const char *mnt, const char *type, mode_t rootmode,
     char *d;
     char *fsname = NULL;
     int check_empty = 1;
+    int blkdev = 0;
 
     optbuf = (char *) malloc(strlen(opts) + 128);
     if (!optbuf) {
@@ -585,6 +627,12 @@ static int do_mount(const char *mnt, const char *type, mode_t rootmode,
             }
             memcpy(fsname, s + fsname_str_len, len - fsname_str_len);
             fsname[len - fsname_str_len] = '\0';
+        } else if (opt_eq(s, len, "blkdev")) {
+            if (getuid() != 0) {
+                fprintf(stderr, "%s: option blkdev is privileged\n", progname);
+                goto err;
+            }
+            blkdev = 1;
         } else if (opt_eq(s, len, "nonempty")) {
             check_empty = 0;
         } else if (!begins_with(s, "fd=") &&
@@ -646,14 +694,20 @@ static int do_mount(const char *mnt, const char *type, mode_t rootmode,
     if (check_empty && check_mountpoint_empty(mnt, rootmode, rootsize) == -1)
         goto err;
 
-    res = mount(fsname, mnt, type, flags, optbuf);
+    if (blkdev)
+        *type = "fuseblk";
+    res = mount(fsname, mnt, *type, flags, optbuf);
     if (res == -1 && errno == EINVAL) {
         /* It could be an old version not supporting group_id */
         sprintf(d, "fd=%i,rootmode=%o,user_id=%i", fd, rootmode, getuid());
-        res = mount(fsname, mnt, type, flags, optbuf);
+        res = mount(fsname, mnt, *type, flags, optbuf);
     }
     if (res == -1) {
-        fprintf(stderr, "%s: mount failed: %s\n", progname, strerror(errno));
+        int errno_save = errno;
+        if (blkdev && errno == ENODEV && !has_fuseblk())
+            fprintf(stderr, "%s: 'fuseblk' support missing; try the kernel module from fuse-2.6.0 or later\n", progname);
+        else
+            fprintf(stderr, "%s: mount failed: %s\n", progname, strerror(errno_save));
         goto err;
     } else {
         *fsnamep = fsname;
@@ -794,7 +848,8 @@ static int try_open(const char *dev, char **devp, int silent)
             close(fd);
             fd = -1;
         }
-    } else if (errno == ENODEV)
+    } else if (errno == ENODEV ||
+               errno == ENOENT) /* check for ENOENT too, for the udev case */
         return -2;
     else if (!silent) {
         fprintf(stderr, "%s: failed to open %s: %s\n", progname, dev,
@@ -825,13 +880,12 @@ static int try_open_fuse_device(char **devp)
 static int open_fuse_device(char **devp)
 {
     int fd = try_open_fuse_device(devp);
-    if (fd >= 0)
+    if (fd >= -1)
         return fd;
 
-    if (fd == -2)
-        fprintf(stderr,
-                "%s: fuse device not found, try 'modprobe fuse' first\n",
-                progname);
+    fprintf(stderr, "%s: fuse device not found, try 'modprobe fuse' first\n",
+            progname);
+
     return -1;
 }
 
@@ -854,13 +908,8 @@ static int mount_fuse(const char *mnt, const char *opts)
     if (fd == -1)
         return -1;
 
-    if (geteuid() == 0) {
+    if (geteuid() == 0)
         mtablock = lock_mtab();
-        if (mtablock < 0) {
-            close(fd);
-            return -1;
-        }
-    }
 
     drop_privs();
     read_conf();
@@ -880,7 +929,7 @@ static int mount_fuse(const char *mnt, const char *opts)
         res = check_perm(&real_mnt, &stbuf, &currdir_fd, &mountpoint_fd);
         restore_privs();
         if (res != -1)
-            res = do_mount(real_mnt, type, stbuf.st_mode & S_IFMT, fd, opts,
+            res = do_mount(real_mnt, &type, stbuf.st_mode & S_IFMT, fd, opts,
                            dev, &fsname, &mnt_opts, stbuf.st_size);
     } else
         restore_privs();
@@ -987,7 +1036,7 @@ static int send_fd(int sock_fd, int fd)
     struct msghdr msg;
     struct cmsghdr *p_cmsg;
     struct iovec vec;
-    char cmsgbuf[CMSG_SPACE(sizeof(fd))];
+    size_t cmsgbuf[CMSG_SPACE(sizeof(fd)) / sizeof(size_t)];
     int *p_fds;
     char sendchar = 0;
 
